@@ -1,5 +1,4 @@
 # -*- coding: utf-8 -*-
-import itertools
 import json
 import logging
 import re
@@ -25,19 +24,8 @@ class Conceptualizer:
         model_path: str = None,
         word_delimiter: str = "_",
         ent_score: bool = False,
-        exclude_pos: list = [
-            "VERB",
-            "AUX",
-            "ADP",
-            "DET",
-            "CCONJ",
-            "PUNCT",
-            "ADV",
-            "ADJ",
-            "PART",
-            "PRON",
-        ],
-        exclude_dep: list = [],
+        exclude_pos: list = None,
+        exclude_dep: list = None,
         include_compound_words: bool = False,
         case_sensitive: bool = False,
         json_path: str = "./matching_patterns.json",
@@ -81,8 +69,23 @@ class Conceptualizer:
         self.topn = topn
         self.model_path = model_path
         self.match_rule = {}
+        if exclude_pos is None:
+            exclude_pos = [
+                "VERB",
+                "AUX",
+                "ADP",
+                "DET",
+                "CCONJ",
+                "PUNCT",
+                "ADV",
+                "ADJ",
+                "PART",
+                "PRON",
+            ]
         if exclude_pos:
             self.match_rule["POS"] = {"NOT_IN": exclude_pos}
+        if exclude_dep is None:
+            exclude_dep = []
         if exclude_dep:
             self.match_rule["DEP"] = {"NOT_IN": exclude_dep}
         self.json_path = json_path
@@ -101,13 +104,12 @@ class Conceptualizer:
         self.run()
         self.data_upper = {k.upper(): v for k, v in data.items()}
 
-    def run(self):
+    def run(self) -> None:
         self.check_validity_path()
         self.determine_topn()
         self.set_gensim_model()
         self.verify_data(self.verbose)
         self.expand_concepts()
-        self.verify_data(verbose=False)
         # settle words around overlapping concepts
         for _ in range(5):
             self.expand_concepts()
@@ -119,7 +121,11 @@ class Conceptualizer:
         if not self.ent_score:
             del self.kv
 
-    def check_validity_path(self):
+    def check_validity_path(self) -> None:
+        """
+        If the path is a file, create the parent directory if it doesn't exist. If the path is a directory, create the
+        directory and set the path to the default file name
+        """
         if self.json_path:
             if Path(self.json_path).suffix:
                 Path(self.json_path).parents[0].mkdir(parents=True, exist_ok=True)
@@ -132,24 +138,21 @@ class Conceptualizer:
                     f" ´json_path´to {self.json_path}"
                 )
 
-    def determine_topn(self):
+    def determine_topn(self) -> None:
         """
         If the user doesn't specify a topn value for each class,
         then the topn value for each class is set to 100
         """
-        self.topn_dict = {}
         if not self.topn:
-            for key in self.data:
-                self.topn_dict[key] = 100
+            self.topn_dict = {key: 100 for key in self.data}
         else:
-            num_classes = len(list(self.data.keys()))
+            num_classes = len(self.data)
             assert (
                 len(self.topn) == num_classes
             ), f"Provide a topn integer for each of the {num_classes} classes."
-            for key, n in zip(self.data, self.topn):
-                self.topn_dict[key] = n
+            self.topn_dict = dict(zip(self.data, self.topn))
 
-    def set_gensim_model(self):
+    def set_gensim_model(self) -> None:
         """
         If the model_path is not None, then we try to load the model from the path.
         If it's not a valid path, then we raise an exception.
@@ -179,7 +182,7 @@ class Conceptualizer:
 
             assert len(
                 self.nlp.vocab.vectors
-            ), "Choose a model with internal embeddings i.e. md or lg."
+            ), "Choose a spaCy model with internal embeddings, e.g. md or lg."
 
             for key, vector in self.nlp.vocab.vectors.items():
                 wordList.append(self.nlp.vocab.strings[key])
@@ -189,107 +192,82 @@ class Conceptualizer:
 
             self.kv.add_vectors(wordList, vectorList)
 
-    def verify_data(self, verbose: bool = True):
+    def verify_data(self, verbose: bool = True) -> None:
         """
         It takes a dictionary of lists of words, and returns a dictionary of lists of words,
         where each word in the list is present in the word2vec model
         """
-        verified_data = dict()
+        verified_data: dict[str, list[str]] = dict()
         for key, value in self.data.items():
             verified_values = []
-            if not self.check_presence_vocab(key):
-                if verbose:
-                    if key not in self.log_cache["key"]:
-                        logger.warning(f"key ´{key}´ not present in vector model")
-                        self.log_cache["key"].append(key)
+            present_key = self.check_presence_vocab(key)
+            if not present_key and verbose and key not in self.log_cache["key"]:
+                logger.warning(f"key ´{key}´ not present in vector model")
+                self.log_cache["key"].append(key)
             for word in value:
-                if self.check_presence_vocab(word):
-                    verified_values.append(self.check_presence_vocab(word))
-                else:
-                    if verbose:
-                        if word not in self.log_cache["word"]:
-                            logger.warning(
-                                f"word ´{word}´ from key ´{key}´ not present in vector"
-                                " model"
-                            )
-                            self.log_cache["word"].append(word)
+                present_word = self.check_presence_vocab(word)
+                if present_word:
+                    verified_values.append(present_word)
+                elif verbose and word not in self.log_cache["word"]:
+                    logger.warning(
+                        f"word ´{word}´ from key ´{key}´ not present in vector model"
+                    )
+                    self.log_cache["word"].append(word)
             verified_data[key] = verified_values
             if not len(verified_values):
                 msg = (
                     f"None of the entries for key {key} are present in the vector"
                     " model. "
                 )
-                if self.check_presence_vocab(key):
-                    logger.warning(msg + f"Using {key} as word to expand over instead.")
-                    verified_data[key] = self.check_presence_vocab(key)
+                if present_key:
+                    logger.warning(
+                        msg + f"Using {present_key} as word to expand over instead."
+                    )
+                    verified_data[key] = present_key
                 else:
                     raise Exception(msg)
         self.data = deepcopy(verified_data)
-        self.original_data = deepcopy(self.data)
 
-    def expand_concepts(self):
+    def expand_concepts(self) -> None:
         """
         For each key in the data dictionary, find the topn most similar words to the key and the values in the data
         dictionary, and add those words to the values in the data dictionary
         """
 
+        self.original_data = deepcopy(self.data)
+
         for key in self.data:
-            remaining_keys = [rem_key for rem_key in self.data.keys() if rem_key != key]
-            remaining_values = [self.data[rem_key] for rem_key in remaining_keys]
-            remaining_values = list(itertools.chain.from_iterable(remaining_values))
-            if self.check_presence_vocab(key):
-                key_list = [self.check_presence_vocab(key)]
+            present_key = self.check_presence_vocab(key)
+            if present_key:
+                key_list = [present_key]
             else:
                 key_list = []
             similar = self.kv.most_similar(
                 positive=self.data[key] + key_list,
                 topn=self.topn_dict[key],
             )
-            similar = [sim_pair[0] for sim_pair in similar]
-            self.data[key] += similar
-            self.data[key] = list(set([word.lower() for word in self.data[key]]))
+            self.data[key] = list(
+                {self.check_presence_vocab(word) for word, _ratio in similar}
+            )
 
-    def resolve_overlapping_concepts(self):
+    def resolve_overlapping_concepts(self) -> None:
         """
         It removes words from the data that are in other concepts, and then removes words that are not closest to the
         centroid of the concept
         """
-        centroids = {}
-        for key in self.data:
-            if not self.check_presence_vocab(key):
-                words = self.data[key]
-                while len(words) != 1:
-                    words.remove(self.kv.doesnt_match(words))
-                centroids[key] = words[0]
-            else:
-                centroids[key] = key
-
-        for key_x in self.data:
-            for key_y in self.data:
-                if key_x != key_y:
-                    self.data[key_x] = [
-                        word
-                        for word in self.data[key_x]
-                        if word not in self.original_data[key_y]
-                    ]
-
         for key in self.data:
             self.data[key] = [
                 word
                 for word in self.data[key]
-                if centroids[key]
-                == self.kv.most_similar_to_given(word, list(centroids.values()))
+                if key == self.kv.most_similar_to_given(word, list(self.data.keys()))
             ]
 
-        self.centroids = centroids
-
-    def infer_original_data(self):
+    def infer_original_data(self) -> None:
         """
         It takes the original data and adds the new data to it, then removes the new data from the original data.
         """
         for key in self.data:
-            self.data[key] += self.original_data[key]
-            self.data[key] = list(set(self.data[key]))
+            self.data[key] = list(set(self.data[key] + self.original_data[key]))
 
         for key_x in self.data:
             for key_y in self.data:
@@ -300,9 +278,7 @@ class Conceptualizer:
                         if word not in self.original_data[key_y]
                     ]
 
-        self.verify_data(verbose=False)
-
-    def lemmatize_concepts(self):
+    def lemmatize_concepts(self) -> None:
         """
         For each key in the data dictionary,
         the function takes the list of concepts associated with that key, and lemmatizes
@@ -313,7 +289,7 @@ class Conceptualizer:
                 set([doc[0].lemma_ for doc in self.nlp.pipe(self.data[key])])
             )
 
-    def create_conceptual_patterns(self):
+    def create_conceptual_patterns(self) -> None:
         """
         For each key in the data dictionary,
         create a pattern for each word in the list of words associated with that key.
@@ -341,13 +317,21 @@ class Conceptualizer:
         """
         patterns = []
 
-        def add_patterns(input_dict):
+        def add_patterns(input_dict: dict) -> None:
+            """
+            It creates a  list of dictionaries that can be used for a spaCy entity ruler
+
+            :param input_dict: a dictionary
+            :type input_dict: dict
+            """
             for key in input_dict:
                 if self.match_key == "LEMMA":
                     words = [
                         "".join(
                             [
-                                token.lemma_ if token.lemma_ else token.text
+                                token.lemma_ + token.whitespace_
+                                if token.lemma_
+                                else token.text
                                 for token in doc
                             ]
                         )
@@ -357,9 +341,10 @@ class Conceptualizer:
                     words = input_dict[key]
                 for word in words:
                     if word != key:
-                        specific_match_rule = dict()
-                        specific_match_rule.update(self.match_rule)
-                        word_parts = re.split(f"[{self.word_delimiter}]+", word)
+                        specific_match_rule = {**self.match_rule}
+                        word_parts = re.split(
+                            f"[{re.escape(self.word_delimiter)}]+", word
+                        )
                         if len(word_parts) > 1:
                             operators = [" ", "-"]
                         else:
@@ -367,7 +352,7 @@ class Conceptualizer:
 
                         for op in operators:
                             if self.case_sensitive:
-                                specific_match_rule[self.match_key] = "{op}".join(
+                                specific_match_rule[self.match_key] = f"{op}".join(
                                     word_parts
                                 )
                             else:
@@ -404,14 +389,13 @@ class Conceptualizer:
                                 )
 
         add_patterns(self.data)
-        add_patterns(self.original_data)
         if self.json_path:
             with open(self.json_path, "w") as f:
                 json.dump(patterns, f)
         self.ruler = self.nlp.add_pipe("entity_ruler", config={"overwrite_ents": True})
         self.ruler.add_patterns(patterns)
 
-    def __call__(self, doc: Doc):
+    def __call__(self, doc: Doc) -> Doc:
         """
         It takes a doc object and assigns a score to each entity in the doc object
 
@@ -422,7 +406,7 @@ class Conceptualizer:
             doc = self.assign_score_to_entities(doc)
         return doc
 
-    def pipe(self, stream, batch_size=128):
+    def pipe(self, stream, batch_size=128) -> Doc:
         """
         It takes a stream of documents, and for each document,
         it assigns a score to each entity in the document
@@ -436,7 +420,7 @@ class Conceptualizer:
                     doc = self.assign_score_to_entities(doc)
                 yield doc
 
-    def assign_score_to_entities(self, doc: Doc):
+    def assign_score_to_entities(self, doc: Doc) -> Doc:
         """
         The function takes a spaCy document as input and assigns a score to each entity in the document. The score is
         calculated using the word embeddings of the entity and the concept.
@@ -453,16 +437,16 @@ class Conceptualizer:
                 if self.check_presence_vocab(ent.text):
                     entity = [ent.text]
                 else:
-                    entity = [
-                        self.check_presence_vocab(part)
-                        for part in ent.text.split()
-                        if self.check_presence_vocab(part)
-                    ]
-                concept = [
-                    self.check_presence_vocab(word)
-                    for word in self.data_upper[ent.label_]
-                    if self.check_presence_vocab(word)
-                ]
+                    entity = []
+                    for part in ent.text.split():
+                        present_part = self.check_presence_vocab(part)
+                        if present_part:
+                            entity.append(present_part)
+                concept = []
+                for word in self.data_upper[ent.label_]:
+                    present_word = self.check_presence_vocab(word)
+                    if present_word:
+                        concept.append(present_word)
                 if entity and concept:
                     ent._.ent_score = self.kv.n_similarity(entity, concept)
                 else:
@@ -487,10 +471,35 @@ class Conceptualizer:
         doc.ents = ents
         return doc
 
-    def check_presence_vocab(self, word):
+    def _check_presence_vocab(self, word: str) -> str:
+        """
+        If the word is in the vocabulary, return the word. If not, replace spaces and dashes with the word delimiter and
+        check if the new word is in the vocabulary. If so, return the new word
+
+        :param word: str - the word to check
+        :type word: str
+        :return: The word or the check_word
+        """
+        if word in self.kv:
+            return word
         for op in [" ", "-"]:
             check_word = word.replace(op, self.word_delimiter)
-            if not self.case_sensitive:
-                check_word = check_word.lower()
             if check_word in self.kv:
                 return check_word
+
+    def check_presence_vocab(self, word: str) -> str:
+        """
+        If the word is not lowercase and the case_sensitive flag is set to False, then check if the lowercase version of
+        the word is in the vocabulary. If it is, return the lowercase version of the word. Otherwise, return the word
+        itself
+
+        :param word: The word to check for presence in the vocabulary
+        :type word: str
+        :return: The word itself if it is present in the vocabulary, otherwise the word with the highest probability of
+        being the word that was intended.
+        """
+        if not word.islower() and not self.case_sensitive:
+            present_word = self._check_presence_vocab(word.lower())
+            if present_word:
+                return present_word
+        return self._check_presence_vocab(word)
